package ru.dbotthepony.kstarbound.defs.world

import com.google.gson.GsonBuilder
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.io.ColorTypeAdapter
import ru.dbotthepony.kstarbound.io.json.CustomEnumTypeAdapter
import ru.dbotthepony.kstarbound.io.json.BuilderAdapter
import ru.dbotthepony.kvector.vector.Color
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty

class SkyParameters {
	var spaceLevel = 3000
	var ambientLightLevel = Color.BLACK
	var skyType = SkyType.ATMOSPHERIC
	var skyColoring = SkyColoring()
	var satellites = Array(0) { SkySatellite() }
	var planet: SkySatellite? = null
	var horizonImages = Array(0) { "" }
	var dayLength = 10_000.0
	var surfaceLevel = 1_200
	var seed = 0L

	companion object {
		val ADAPTER = BuilderAdapter(::SkyParameters,
			SkyParameters::spaceLevel,
			SkyParameters::ambientLightLevel,
			SkyParameters::skyType,
			SkyParameters::skyColoring,
			SkyParameters::satellites,
			SkyParameters::planet,
			SkyParameters::horizonImages,
			SkyParameters::dayLength,
			SkyParameters::surfaceLevel,
			SkyParameters::seed,
		)

		fun registerGson(gsonBuilder: GsonBuilder) {
			gsonBuilder.registerTypeAdapter(SkyParameters::class.java, ADAPTER)
			gsonBuilder.registerTypeAdapter(SkyColoringManifold::class.java, SkyColoringManifold.ADAPTER)
			gsonBuilder.registerTypeAdapter(SkyColoring::class.java, SkyColoring.ADAPTER)
			gsonBuilder.registerTypeAdapter(SkyType::class.java, CustomEnumTypeAdapter(SkyType.values()).nullSafe())
			SkySatellite.registerGson(gsonBuilder)
		}
	}
}

class SkyColoringManifold {
	var initial = Color.TRANSLUCENT
	var final = Color.TRANSLUCENT

	companion object {
		val ADAPTER = object : TypeAdapter<SkyColoringManifold>() {
			override fun write(out: JsonWriter?, value: SkyColoringManifold?) {
				TODO("Not yet implemented")
			}

			override fun read(reader: JsonReader): SkyColoringManifold {
				val instance = SkyColoringManifold()
				reader.beginArray()
				instance.initial = ColorTypeAdapter.read(reader)
				instance.final = ColorTypeAdapter.read(reader)
				reader.endArray()
				return instance
			}
		}
	}
}

class SkyColoring {
	var mainColor = Color.TRANSLUCENT

	var morningColors = SkyColoringManifold()
	var dayColors = SkyColoringManifold()
	var eveningColors = SkyColoringManifold()
	var nightColors = SkyColoringManifold()

	var morningLightColor = Color.TRANSLUCENT
	var dayLightColor = Color.TRANSLUCENT
	var eveningLightColor = Color.TRANSLUCENT
	var nightLightColor = Color.TRANSLUCENT

	companion object {
		val ADAPTER = BuilderAdapter(::SkyColoring,
			SkyColoring::mainColor,
			SkyColoring::morningColors,
			SkyColoring::dayColors,
			SkyColoring::eveningColors,
			SkyColoring::nightColors,
			SkyColoring::morningLightColor,
			SkyColoring::dayLightColor,
			SkyColoring::eveningLightColor,
			SkyColoring::nightLightColor,
		)
	}
}

class SkySatellite {
	class Layer {
		private inner class ImageDelegate : ReadWriteProperty<Layer, String> {
			private var value: String? = null

			override fun getValue(thisRef: Layer, property: KProperty<*>): String {
				return value ?: throw IllegalStateException("Property is not initialized")
			}

			override fun setValue(thisRef: Layer, property: KProperty<*>, value: String) {
				val split = value.split('?')
				this.value = split[0]

				for (prop in 1 until split.size) {
					val splitProp = split[prop].split('=')

					if (splitProp.size == 1) {
						throw IllegalArgumentException("Invalid property: ${split[prop]}, near $prop, in $value")
					}

					when (val propName = splitProp[0]) {
						"hueshift" -> hueShift = splitProp[1].toInt()
						"addmask" -> mask = splitProp[1]
						else -> LOGGER.warn("Unknown property $propName near $prop in $value")
					}
				}
			}
		}

		var scale = 1.0
		var image by ImageDelegate()
		var hueShift = 0
		var mask: String? = null
	}

	var pos = Vector2d.ZERO
	var layers = Array(0) { Layer() }

	companion object {
		val ADAPTER = BuilderAdapter(::SkySatellite,
			SkySatellite::pos,
			SkySatellite::layers,
		)

		val LAYER_ADAPTER = BuilderAdapter(::Layer,
			Layer::scale,
			Layer::image,
			Layer::hueShift,
			Layer::mask,
		)

		private val LOGGER = LogManager.getLogger(SkySatellite::class.java)

		fun registerGson(gsonBuilder: GsonBuilder) {
			gsonBuilder.registerTypeAdapter(SkySatellite::class.java, ADAPTER)
			gsonBuilder.registerTypeAdapter(Layer::class.java, LAYER_ADAPTER)
		}
	}
}
